# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Query Tox for the tags defined in the specified file."""

# mypy needs these assertions, and they are better expressed in a compact manner

from __future__ import annotations

import ast
import configparser
import itertools
import pathlib
import subprocess  # noqa: S404
import sys
from typing import Final, NamedTuple

from tox_trivtags import HAVE_MOD_TOX_3


DEFAULT_FILENAME = pathlib.Path("tox.ini")


class TestenvTags(NamedTuple):
    """A Tox environment along with its tags."""

    cfg_name: str
    name: str
    tags: list[str]


def _validate_parsed_bool(value: ast.expr) -> bool:
    """Make sure a boolean value is indeed a boolean value."""
    assert isinstance(value, ast.Constant) and isinstance(value.value, bool)
    return value.value


def _validate_parsed_str(value: ast.expr) -> str:
    """Make sure a string is indeed a string."""
    assert isinstance(value, ast.Constant) and isinstance(value.value, str)
    return value.value


def _validate_parsed_strlist(value: ast.expr) -> list[str]:
    """Make sure a list of strings is indeed a list of strings."""
    assert isinstance(value, ast.List)
    return [_validate_parsed_str(value) for value in value.elts]


def _parse_bool(value: str) -> bool:
    """Parse a Python-esque representation of a boolean value without eval()."""
    a_body: Final = ast.parse(value).body
    assert len(a_body) == 1 and isinstance(a_body[0], ast.Expr)
    return _validate_parsed_bool(a_body[0].value)


def _parse_strlist(value: str) -> list[str]:
    """Parse a Python-esque representation of a list of strings without eval()."""
    a_body: Final = ast.parse(value).body
    assert len(a_body) == 1 and isinstance(a_body[0], ast.Expr)
    return _validate_parsed_strlist(a_body[0].value)


def _parse_lines(value: str) -> list[str]:
    """Parse a list of text lines as Tox 4 wants to output the tags."""
    return [line for line in value.splitlines() if line]


def _parse_tags(value: str) -> list[str]:
    """Invoke `_parse_lines()` or `_parse_strlist()` as appropriate."""
    return _parse_strlist(value) if value.lstrip().startswith("[") else _parse_lines(value)


def remove_prefix(value: str, prefix: str) -> str:
    """Remove a string's prefix if it is there.

    Will be replaced with str.removeprefix() once we can depend on Python 3.9+.
    """
    parts: Final = value.partition(prefix)
    return parts[2] if parts[1] and not parts[0] else value


def parse_showconfig(
    filename: pathlib.Path = DEFAULT_FILENAME,
    *,
    env: dict[str, str] | None = None,
    tox_invoke: list[str | pathlib.Path] | None = None,
) -> dict[str, TestenvTags]:
    """Run `tox --showconfig` and look for tags in its output."""
    if tox_invoke is None:
        tox_invoke = [sys.executable, "-u", "-m", "tox"]
    tox_cmd: Final = [
        *tox_invoke,
        "--showconfig" if HAVE_MOD_TOX_3 else "config",
        "-c",
        filename,
        "-e",
        "ALL",
    ]

    def parse_output() -> configparser.ConfigParser:
        """Run Tox, parse its output as an INI-style file."""
        lines: Final = subprocess.run(
            tox_cmd,
            check=True,
            encoding="UTF-8",
            env=env,
            shell=False,  # noqa: S603
            stdout=subprocess.PIPE,
        ).stdout.splitlines()
        # Drop the lines that Tox outputs at the start if it needs to bootstrap
        # a more recent version due to a `tox.minversion` specification.
        lines_real: Final = itertools.dropwhile(lambda line: not line.startswith("["), lines)

        cfgp: Final = configparser.ConfigParser(interpolation=None)
        cfgp.read_string("\n".join(lines_real) + "\n")
        return cfgp

    def process_config(cfgp: configparser.ConfigParser) -> dict[str, TestenvTags]:
        """Build the result dictionary."""
        return {
            name: TestenvTags(
                cfg_name=cfg_name,
                name=name,
                tags=_parse_tags(tags),
            )
            for cfg_name, name, tags in (
                (cfg_name, name, env["tags"])
                for cfg_name, name, env in (
                    (cfg_name, remove_prefix(cfg_name, "testenv:"), env)
                    for cfg_name, env in cfgp.items()
                )
                if cfg_name != name
            )
        }

    cfgp_no_req: Final = parse_output()
    if any("tags" in env for env in cfgp_no_req.values()):
        return process_config(cfgp_no_req)

    tox_cmd.extend(["-x", "tox.requires=test-stages >= 0.1.3"])
    return process_config(parse_output())



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""The main tox-stages command-line executable."""

from __future__ import annotations

import dataclasses
import subprocess  # noqa: S404
import sys
from typing import TYPE_CHECKING

import tox_trivtags
from test_stages import cmd


if tox_trivtags.HAVE_MOD_TOX_3 or tox_trivtags.HAVE_MOD_TOX_4:
    from tox_trivtags import parse as ttt_parse

if TYPE_CHECKING:
    import pathlib
    from typing import Final


@dataclasses.dataclass(frozen=True)
class Config(cmd.Config):
    """Also store the path to the Tox executable if found."""

    tox_program: list[str | pathlib.Path] | None = None


@cmd.click_available()
def _cmd_available(cfg: cmd.Config) -> bool:
    """Check whether we can parse the Tox configuration in any of the supported ways.

    Currently the only supported way is `tox --showconfig`.
    """
    assert isinstance(cfg, Config)  # noqa: S101  # mypy needs this
    return cfg.tox_program is not None


@cmd.click_run()
def _cmd_run(cfg: cmd.Config, stages: list[cmd.TestStage], extra_args: list[str]) -> None:
    """Run the Tox environments in groups."""
    toxdir = cfg.filename.parent

    def run_group(group: list[cmd.TestEnv], *, parallel: bool) -> None:
        """Run the stages in a single group."""
        if not isinstance(cfg, Config) or cfg.tox_program is None:
            #  _tox_get_envs() really should have taken care of that
            sys.exit(f"Internal error: tox-stages run_group: Config? {cfg!r}")

        names: Final = ",".join(env.name for env in group)
        print(f"\n=== Running Tox environments: {names}\n")  # noqa: T201
        run_parallel = (
            (["-p", "all"] if parallel else [])
            if tox_trivtags.HAVE_MOD_TOX_3
            else (["run-parallel"] if parallel else ["run"])
        )
        tox_cmd: Final = [*cfg.tox_program, *run_parallel, "-e", names, *extra_args]
        res: Final = subprocess.run(
            tox_cmd,
            check=False,
            cwd=toxdir,
            env=cfg.utf8_env,
            shell=False,  # noqa: S603
        )
        if res.returncode != 0:
            sys.exit(f"Tox failed for the {names} environments")

    for desc in stages:
        run_group(desc.envlist, parallel=desc.parallel)

    print("\n=== All Tox environment groups passed!")  # noqa: T201


def _tox_get_envs(cfg: cmd.Config) -> list[cmd.TestEnv]:
    """Get all the Tox environments from the config file."""
    assert isinstance(cfg, Config)  # noqa: S101  # mypy needs this
    if cfg.tox_program is None:
        sys.exit("No tox program found or specified")
    tcfg: Final = ttt_parse.parse_showconfig(
        filename=cfg.filename,
        env=cfg.utf8_env,
        tox_invoke=cfg.tox_program,
    )
    return [cmd.TestEnv(name, env.tags) for name, env in tcfg.items()]


def _find_tox_program() -> list[str | pathlib.Path] | None:
    """Figure out how to invoke Tox.

    For the present, only a Tox installation in the current Python interpreter's
    package directories is supported, since we need to be sure that we can rely on
    the `tox-trivtags` package being installed.

    Also, we only support Tox 3.x for the present.
    """
    if not tox_trivtags.HAVE_MOD_TOX_3 and not tox_trivtags.HAVE_MOD_TOX_4:
        return None

    return [sys.executable, "-m", "tox"]


@cmd.click_main(
    prog="tox-stages",
    prog_help="Run Tox environments in groups, stop on failure.",
    filename="tox.ini",
    filename_help="the path to the Tox config file to parse",
    get_all_envs=_tox_get_envs,
)
def main(cfg: cmd.Config) -> cmd.Config:
    """Return our `Config` object with the path to Tox if found."""
    return Config(**dataclasses.asdict(cfg), tox_program=_find_tox_program())


main.add_command(_cmd_available)
main.add_command(_cmd_run)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Run a test for the `test-stages` library's `tox-stages` runner."""

from __future__ import annotations

import os
import pathlib
import subprocess
import sys
import tarfile
import tempfile
import typing

import pyproject_hooks
import tomli_w
import utf8_locale


if sys.version_info >= (3, 11):
    import contextlib as contextlib_chdir

    import tomllib
else:
    import contextlib_chdir
    import tomli as tomllib


if typing.TYPE_CHECKING:
    from typing import Final


def validate_srcdir(srcdir: pathlib.Path) -> None:
    """Make sure we can find a couple of files in the source directory."""
    for relpath in (
        "requirements/install.txt",
        "src/test_stages/tox_stages/__main__.py",
        "tests/unit/test_functional.py",
    ):
        path = srcdir / relpath
        if not path.is_file():
            sys.exit(f"Expected to find {relpath} in {srcdir}, but {path} is not a regular file")


def build_sdist(srcdir: pathlib.Path, tempd: pathlib.Path) -> pathlib.Path:
    """Build a source distribution tarball."""
    with contextlib_chdir.chdir(tempd):
        distdir: Final = tempd / "dist"
        distdir.mkdir(mode=0o755)

        backend: Final = tomllib.loads((srcdir / "pyproject.toml").read_text(encoding="UTF-8"))[
            "build-system"
        ]["build-backend"]

        caller: Final = pyproject_hooks.BuildBackendHookCaller(srcdir, backend)
        fname: Final = caller.build_sdist(distdir)
        sdist: Final = distdir / fname
        if sdist.parent != distdir:
            sys.exit(f"The PEP517 build returned {fname} which does not seem to be a pure filename")
        return sdist


def safe_extract_all(star: tarfile.TarFile, topdir: pathlib.Path) -> None:
    """Validate the member names in the archive and extract them all."""
    members: Final = star.getmembers()
    paths: Final = [pathlib.Path(member.name) for member in members]
    if not paths:
        sys.exit("Expected at least one member in the source archive")
    if paths[0].is_absolute():
        sys.exit(f"Did not expect an absolute path {paths[0]} in the source archive")

    base_path: Final = paths[0].parts[0]
    bad_paths: Final = [path for path in paths if path.parts[0] != base_path]
    if bad_paths:
        sys.exit(
            f"Bad paths in the source archive, expected all of them to "
            f"start with {base_path}: {bad_paths}",
        )

    bad_dirs: Final = [path for path in paths if ".." in path.parts]
    if bad_dirs:
        sys.exit(f"Bad paths in the source archive, none of them should contain '..': {bad_paths}")

    star.extractall(topdir, members=members)  # noqa: S202


def extract_sdist(sdist: pathlib.Path, tempd: pathlib.Path) -> pathlib.Path:
    """Extract the sdist tarball."""
    if not sdist.name.endswith(".tar.gz"):
        sys.exit(f"The PEP517 build generated a non-.tar.gz file: {sdist}")

    topdir: Final = tempd / "src"
    topdir.mkdir(mode=0o755)

    with tarfile.open(sdist, mode="r") as star:
        safe_extract_all(star, topdir)
        entries: Final = sorted(path for path in topdir.iterdir())
        if len(entries) != 1:
            sys.exit(f"Expected {sdist} to contain a single directory, got {entries!r}")

    testdir: Final = entries[0]
    if not testdir.is_dir() or not testdir.name.startswith(("test_stages-", "test-stages-")):
        sys.exit(f"Expected {sdist} to contain a single `test-stages-*` directory, got {testdir}")
    return testdir


def adapt_pyproject(testdir: pathlib.Path) -> None:
    """Disable this selftest to avoid infinite recursion."""
    projfile: Final = testdir / "pyproject.toml"
    projdata: Final = tomllib.loads(projfile.read_text(encoding="UTF-8"))

    stages: Final = projdata["tool"]["test-stages"]["stages"]
    if not stages[-1].startswith("@tests"):
        sys.exit(f"Expected a `@tests...` test-stages entry, got {stages!r}")
    stages[-1] = f"{stages[-1]} and not selftest"

    projfile.write_text(tomli_w.dumps(projdata), encoding="UTF-8")


def run_tox(testdir: pathlib.Path) -> None:
    """Clean up the environment a bit, then run Tox."""
    env: Final = dict(item for item in os.environ.items() if not item[0].startswith("TOX"))
    subprocess.check_call(["pwd"], cwd=testdir, env=env)
    subprocess.check_call(["cat", "pyproject.toml"], cwd=testdir, env=env)

    subprocess.check_call(["tox-stages", "available"], cwd=testdir, env=env)

    marker: Final = testdir / "selftest-marker.txt"
    if marker.is_symlink() or marker.exists():
        sys.exit(f"Did not expect {marker} to exist")
    subprocess.check_call(
        ["python3", "-m", "test_stages.tox_stages", "run", "@selftest"],
        cwd=testdir,
        env=env,
    )
    if not marker.is_file():
        sys.exit(f"`tox-stages run @selftest` did not create {marker}")

    marker.unlink()
    subprocess.check_call(
        ["python3", "-m", "test_stages.tox_stages", "run", "--arg", "--notest", "@selftest"],
        cwd=testdir,
        env=env,
    )
    if marker.is_symlink() or marker.exists():
        sys.exit(f"A `--notest` run still created {marker}")

    subprocess.check_call(
        ["python3", "-m", "test_stages.tox_stages", "run", "(@docs or not @manual) and @selftest"],
        cwd=testdir,
        env=env,
    )
    if not marker.is_file():
        sys.exit(f"`tox-stages run (@docs or not @manual) and @selftest` did not create {marker}")

    marker.unlink()
    subprocess.check_call(
        ["python3", "-m", "test_stages.tox_stages", "run", "-m", "@selftest", "not @manual"],
        cwd=testdir,
        env=env,
    )
    if not marker.is_file():
        sys.exit(f"`tox-stages run -m @selftest not @manual` did not create {marker}")

    utf8_env = dict(env)
    utf8_env.update(utf8_locale.UTF8Detect().detect().env_vars)
    blurb = "import pathlib"
    if blurb in subprocess.check_output(
        ["tox-stages", "run", "@selftest"],
        cwd=testdir,
        encoding="UTF-8",
        env=utf8_env,
    ):
        sys.exit("A run without any -p option output {blurb!r}")

    if blurb in subprocess.check_output(
        ["tox-stages", "run", "@selftest", "-p", "1"],
        cwd=testdir,
        encoding="UTF-8",
        env=utf8_env,
    ):
        sys.exit("A `-p 1` run did not output {blurb!r}")

    if blurb not in subprocess.check_output(
        ["tox-stages", "run", "@selftest", "-p", "7"],
        cwd=testdir,
        encoding="UTF-8",
        env=utf8_env,
    ):
        sys.exit("A `-p 7` run output {blurb!r}")

    subprocess.check_call(["tox-stages", "run"], cwd=testdir, env=env)


def main() -> None:
    """Build a source distribution, extract it, run some tests."""
    srcdir: Final = pathlib.Path.cwd()
    validate_srcdir(srcdir)

    with tempfile.TemporaryDirectory() as tempd_name:
        tempd: Final = pathlib.Path(tempd_name)
        sdist: Final = build_sdist(srcdir, tempd)
        testdir: Final = extract_sdist(sdist, tempd)
        adapt_pyproject(testdir)
        run_tox(testdir)


if __name__ == "__main__":
    main()



from _typeshed import StrOrBytesPath
from typing import Generic, TypeVar
from contextlib import AbstractContextManager

_T_fd_or_any_path = TypeVar("_T_fd_or_any_path", bound=int | StrOrBytesPath)

class chdir(AbstractContextManager[None], Generic[_T_fd_or_any_path]):
    path: _T_fd_or_any_path
    def __init__(self, path: _T_fd_or_any_path) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, *excinfo: object) -> None: ...



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Create a virtual environment, install Tox, run it."""

from __future__ import annotations

import argparse
import configparser
import dataclasses
import functools
import json
import logging
import pathlib
import shlex
import subprocess
import sys
import tempfile
import typing
import venv


if typing.TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Final


VERSION: Final = "0.1.2"
"""The vetox library version."""


TOX_MIN_VERSION: Final = "4.1"
"""The minimum version of Tox needed to run our tests."""


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for the venv-tox tool."""

    conf: pathlib.Path
    """The path to the `tox.ini` file to use."""

    log: logging.Logger
    """The logger to send diagnostic, informational, warning, and error messages to."""

    tempd: pathlib.Path
    """The temporary directory to operate in."""

    tox_req: str | None
    """The PEP508 version requirements for Tox itself if specified."""


# Shamelessly stolen from the logging-std module
@functools.lru_cache
def build_logger() -> logging.Logger:
    """Build a logger object, send info messages to stdout, everything else to stderr."""
    logger: Final = logging.getLogger("vetox")
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    h_out: Final = logging.StreamHandler(sys.stdout)
    h_out.setLevel(logging.INFO)
    h_out.addFilter(lambda rec: rec.levelno == logging.INFO)
    logger.addHandler(h_out)

    h_err: Final = logging.StreamHandler(sys.stderr)
    h_err.setLevel(logging.INFO)
    h_err.addFilter(lambda rec: rec.levelno != logging.INFO)
    logger.addHandler(h_err)

    return logger


def create_and_update_venv(cfg: Config) -> pathlib.Path:
    """Create a virtual environment, update all the packages within."""
    penv: pathlib.Path = cfg.tempd / "venv"
    cfg.log.info("About to create the %(penv)s virtual environment", {"penv": penv})
    if sys.version_info >= (3, 9):
        cfg.log.info("- using venv.create(upgrade_deps) directly")
        venv.create(penv, with_pip=True, upgrade_deps=True)
        return penv

    cfg.log.info("- no venv.create(upgrade_deps)")
    venv.create(penv, with_pip=True)

    cfg.log.info("- obtaining the list of packages in the virtual environment")
    contents: Final = subprocess.check_output(
        [penv / "bin/python3", "-m", "pip", "list", "--format=json"],
        encoding="UTF-8",
    )
    pkgs: Final = json.loads(contents)
    if (
        not isinstance(pkgs, list)
        or not pkgs
        or not all(isinstance(pkg, dict) and "name" in pkg for pkg in pkgs)
    ):
        sys.exit(f"Unexpected `pip list --format=json` output: {pkgs!r}")

    names: Final = sorted(pkg["name"] for pkg in pkgs)
    cfg.log.info(
        "- upgrading the %(names)s package%(plu)s in the virtual environment",
        {"names": ", ".join(names), "plu": "" if len(names) == 1 else "s"},
    )
    subprocess.check_call([penv / "bin/python3", "-m", "pip", "install", "-U", "--", *names])
    return penv


@functools.lru_cache
def get_tox_min_version(cfg: Config) -> str:
    """Look for a minimum Tox version in the tox.ini file, fall back to TOX_MIN_VERSION."""
    cfgp: Final = configparser.ConfigParser(interpolation=None)
    with cfg.conf.open(encoding="UTF-8") as tox_ini:
        cfgp.read_file(tox_ini)

    return cfgp["tox"].get("min_version", cfgp["tox"].get("minversion", TOX_MIN_VERSION))


def install_tox(cfg: Config, penv: pathlib.Path) -> None:
    """Install Tox into the virtual environment."""
    if cfg.tox_req is not None:
        tox_req = f"tox {cfg.tox_req}"
    else:
        minver: Final = get_tox_min_version(cfg)
        tox_req = f"tox >= {minver}"

    cfg.log.info("Installing Tox %(tox_req)s", {"tox_req": tox_req})
    subprocess.check_call([penv / "bin/python3", "-m", "pip", "install", tox_req])


def get_tox_cmdline(
    cfg: Config,
    penv: pathlib.Path,
    *,
    parallel: bool = True,
    args: list[str],
) -> list[pathlib.Path | str]:
    """Get the Tox command with arguments."""
    penv_py3: Final = penv / "bin/python3"

    def get_run_command() -> list[str]:
        """Get the appropriate command to run Tox in parallel or not."""
        if not parallel:
            return ["run"]

        vers: Final = subprocess.check_output(
            [penv_py3, "-m", "tox", "--version"],
            encoding="UTF-8",
        )
        if vers.startswith("3"):
            return ["run", "-p", "all"]

        return ["run-parallel"]

    cfg.log.info(
        "Running Tox%(parallel)s with %(args)s",
        {
            "parallel": " in parallel" if parallel else "",
            "args": ("additional arguments: " + shlex.join(args))
            if args
            else "no additional arguments",
        },
    )
    run_cmd: Final = get_run_command()
    return [penv_py3, "-m", "tox", "-c", cfg.conf, *run_cmd, *args]


def run_tox(cfg: Config, penv: pathlib.Path, *, parallel: bool = True, args: list[str]) -> None:
    """Run Tox from the virtual environment."""
    subprocess.check_call(get_tox_cmdline(cfg, penv, parallel=parallel, args=args))


def run(cfg_no_tempd: Config, *, parallel: bool, args: list[str]) -> None:
    """Create the virtual environment, install Tox, run it."""
    with tempfile.TemporaryDirectory() as tempd_obj:
        cfg: Final = dataclasses.replace(cfg_no_tempd, tempd=pathlib.Path(tempd_obj))
        penv: Final = create_and_update_venv(cfg)
        install_tox(cfg, penv)
        run_tox(cfg, penv, parallel=parallel, args=args)


def cmd_run(cfg_no_tempd: Config, args: list[str]) -> None:
    """Run the Tox tests sequentially."""
    run(cfg_no_tempd, parallel=False, args=args)


def cmd_run_parallel(cfg_no_tempd: Config, args: list[str]) -> None:
    """Run the Tox tests in parallel."""
    run(cfg_no_tempd, parallel=True, args=args)


def cmd_features(_cfg_no_tempd: Config, _args: list[str]) -> None:
    """Display the list of features supported by the program."""
    print(f"Features: vetox={VERSION} tox=0.1 tox-parallel=0.1")


def cmd_version(_cfg_no_tempd: Config, _args: list[str]) -> None:
    """Display the vetox version."""
    print(f"vetox {VERSION}")


def parse_args() -> tuple[Config, Callable[[Config, list[str]], None], list[str]]:
    """Parse the command-line arguments."""
    parser: Final = argparse.ArgumentParser(prog="vetox")
    parser.add_argument(
        "-c",
        "--conf",
        type=pathlib.Path,
        default=pathlib.Path.cwd() / "tox.ini",
        help="The path to the tox.ini file",
    )

    subp: Final = parser.add_subparsers()
    p_run: Final = subp.add_parser("run", help="Run tests sequentially")
    p_run.add_argument(
        "-t",
        "--tox-req",
        type=str,
        help="specify the PEP508 version requirement for Tox itself",
    )
    p_run.add_argument("args", type=str, nargs="*", help="Additional arguments to pass to Tox")
    p_run.set_defaults(func=cmd_run)

    p_run_p: Final = subp.add_parser("run-parallel", help="Run tests in parallel")
    p_run_p.add_argument(
        "-t",
        "--tox-req",
        type=str,
        help="specify the PEP508 version requirement for Tox itself",
    )
    p_run_p.add_argument("args", type=str, nargs="*", help="Additional arguments to pass to Tox")
    p_run_p.set_defaults(func=cmd_run_parallel)

    p_features: Final = subp.add_parser("features", help="Display the supported program features")
    p_features.set_defaults(func=cmd_features)

    p_version: Final = subp.add_parser("version", help="Display the vetox version")
    p_version.set_defaults(func=cmd_version)

    args: Final = parser.parse_args()

    func: Final[Callable[[Config, list[str]], None] | None] = getattr(args, "func", None)
    if func is None:
        sys.exit("No subcommand specified; use `--help` for a list")

    return (
        Config(
            conf=args.conf,
            log=build_logger(),
            tempd=pathlib.Path("/nonexistent"),
            tox_req=args.tox_req,
        ),
        func,
        getattr(args, "args", []),
    )


def main() -> None:
    """Parse command-line arguments, create a virtual environment, run Tox."""
    cfg_no_tempd, func, args = parse_args()
    func(cfg_no_tempd, args)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""A `test-stages` implementation for the Tox test runner.

This module contains the configuration parsing and runtime glue to
let Tox run its test environments grouped in stages.
"""



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

from collections.abc import Iterable
from typing import Any, Dict, List


class Parser:
    def add_testenv_attribute(
        self,
        name: str,
        type: str,
        help: str,
        default: Any = None,
        postprocess: Any = None,
    ) -> None: ...


class TestenvConfig:
    tags: List[str]


class Config:
    envconfigs: Dict[str, TestenvConfig]


def parseconfig(args: List[str], plugins: Iterable[str] = ()) -> Config: ...



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Command-line tool helpers for the various test-stages implementations."""

from __future__ import annotations

import dataclasses
import functools
import pathlib
import sys
from typing import TYPE_CHECKING, NamedTuple

import click
import parse_stages as parse
import utf8_locale


if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Final, TypeVar

    _T = TypeVar("_T")

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


TestEnv = parse.TaggedFrozen


class Stage(NamedTuple):
    """A stage specification and its boolean expression."""

    spec: str
    expr: parse.BoolExpr
    parallel: bool


class TestStage(NamedTuple):
    """A final representation of a test stage: the environments to run and some attributes."""

    envlist: list[TestEnv]
    """The list of environments to run at this stage."""

    parallel: bool
    """Run the environments in parallel."""


class StagesList:
    """Parse the `--parallel` command-line option argument into a set."""

    stages: set[int]
    """The selected stages, 1-based."""

    def __init__(self, value: str) -> None:
        """Record the selected set of stages."""
        self.stages = set(parse.parse_stage_ids(value))


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for the test runner tool."""

    filename: pathlib.Path
    get_all_envs: Callable[[Config], list[TestEnv]]
    match_spec: parse.BoolExpr | None = None
    stages: list[Stage] = dataclasses.field(default_factory=list)
    utf8_env: dict[str, str] = dataclasses.field(
        default_factory=lambda: utf8_locale.UTF8Detect().detect().env,
    )


@dataclasses.dataclass
class ConfigHolder:
    """Hold a Config object."""

    cfg: Config | None = None


def _split_by(current: list[_T], func: Callable[[_T], bool]) -> tuple[list[_T], list[_T]]:
    """Split an ordered list of items in two by the given predicate."""
    res: Final[tuple[list[_T], list[_T]]] = ([], [])
    for stage in current:
        if func(stage):
            res[1].append(stage)
        else:
            res[0].append(stage)
    return res


def select_stages(cfg: Config, all_stages: list[TestEnv]) -> list[TestStage]:
    """Group the stages as specified."""

    def process_stage(
        acc: tuple[list[TestStage], list[TestEnv]],
        stage: Stage,
    ) -> tuple[list[TestStage], list[TestEnv]]:
        """Stash the environments matched by a stage specification."""
        res, current = acc
        if not current:
            sys.exit(f"No test environments left for {stage.spec}")
        left, matched = _split_by(current, stage.expr.evaluate)
        if not matched:
            sys.exit(f"No test environments matched by {stage.spec}")
        res.append(TestStage(envlist=matched, parallel=stage.parallel))
        return res, left

    res_init: Final[list[TestStage]] = []
    selected: Final = functools.reduce(process_stage, cfg.stages, (res_init, list(all_stages)))[0]
    match_spec: Final = cfg.match_spec
    if match_spec is None:
        return selected

    matched_all: Final = [
        stage._replace(envlist=[env for env in stage.envlist if match_spec.evaluate(env)])
        for stage in selected
    ]
    matched: Final = [stage for stage in matched_all if stage.envlist]
    if not matched:
        sys.exit("None of the selected environments satisfied the additional match condition")
    return matched


def extract_cfg(ctx: click.Context) -> Config:
    """Extract the Config object from the ConfigHolder."""
    cfg_hold: Final = ctx.find_object(ConfigHolder)
    # mypy needs these assertions
    assert cfg_hold is not None  # noqa: S101
    cfg: Final = cfg_hold.cfg
    assert cfg is not None  # noqa: S101
    return cfg


def _find_and_load_pyproject(startdir: pathlib.Path) -> dict[str, Any]:
    """Look for a pyproject.toml file, load it if found."""

    def _find_and_load(path: pathlib.Path) -> dict[str, Any] | None:
        """Check for a pyproject.toml file in the specified directory."""
        proj_file: Final = path / "pyproject.toml"
        if not proj_file.is_file():
            return None

        return tomllib.loads(proj_file.read_text(encoding="UTF-8"))

    # Maybe we should look in the parent directories, too... later.
    for path in (startdir,):
        found = _find_and_load(path)
        if found is not None:
            return found

    # No pyproject.toml file found, nothing to parse
    return {}


def click_available() -> Callable[[Callable[[Config], bool]], click.Command]:
    """Wrap an available() function, checking whether the test runner can be invoked."""

    def inner(handler: Callable[[Config], bool]) -> click.Command:
        """Wrap the available check function."""

        @click.command(name="available")
        @click.pass_context
        def real_available(ctx: click.Context) -> None:
            """Check whether the test runner is available."""
            sys.exit(0 if handler(extract_cfg(ctx)) else 1)

        return real_available

    return inner


def click_run() -> Callable[[Callable[[Config, list[TestStage], list[str]], None]], click.Command]:
    """Wrap a run() function, preparing the configuration."""

    def inner(handler: Callable[[Config, list[TestStage], list[str]], None]) -> click.Command:
        """Wrap the run function."""

        @click.command(name="run")
        @click.option(
            "-A",
            "--arg",
            type=str,
            multiple=True,
            help=(
                "an additional argument to pass to the test runner; "
                "may be specified multiple times"
            ),
        )
        @click.option(
            "-m",
            "--match-spec",
            type=str,
            help="additional stage specifications for the tests to run",
        )
        @click.option(
            "-p",
            "--parallel",
            type=StagesList,
            help="specify which stages to run in parallel (e.g. '1,4-6')",
        )
        @click.argument("stages_spec", nargs=-1, required=False, type=str)
        @click.pass_context
        def real_run(
            ctx: click.Context,
            arg: list[str],
            match_spec: str | None,
            parallel: StagesList | None,
            stages_spec: list[str],
        ) -> None:
            """Run the test environments in stages."""
            cfg_base: Final = extract_cfg(ctx)
            if not stages_spec:
                pyproj: Final = _find_and_load_pyproject(cfg_base.filename.parent)
                stages_spec = pyproj.get("tool", {}).get("test-stages", {}).get("stages", [])
                if not stages_spec:
                    sys.exit("No stages specified either on the command line or in pyproject.toml")

            pstages: Final = set(range(len(stages_spec))) if parallel is None else parallel.stages
            cfg: Final = dataclasses.replace(
                cfg_base,
                match_spec=parse.parse_spec(match_spec) if match_spec is not None else None,
                stages=[
                    Stage(spec, parse.parse_spec(spec), idx in pstages)
                    for idx, spec in enumerate(stages_spec)
                ],
            )
            ctx.obj.cfg = cfg

            handler(cfg, select_stages(cfg, cfg.get_all_envs(cfg)), arg)

        return real_run

    return inner


def click_main(
    prog: str,
    prog_help: str,
    filename: str,
    filename_help: str,
    get_all_envs: Callable[[Config], list[TestEnv]],
) -> Callable[[Callable[[Config], Config]], click.Group]:
    """Wrap a main() function, parsing the top-level options."""

    def inner(main: Callable[[Config], Config]) -> click.Group:
        """Wrap the main function."""

        @click.group(name=prog, help=prog_help)
        @click.option(
            "-f",
            "--filename",
            type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path),
            default=filename,
            help=filename_help,
        )
        @click.pass_context
        def real_main(ctx: click.Context, filename: pathlib.Path) -> None:
            """Run Tox environments in groups, stop on failure."""
            ctx.ensure_object(ConfigHolder)
            ctx.obj.cfg = main(Config(filename=filename, get_all_envs=get_all_envs))

        return real_main

    return inner



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

from collections.abc import Callable
from typing import TypeVar

import tox.config


TParserHook = Callable[[tox.config.Parser], None]


# This only handles the parser hook right now.
def hookimpl(func: TParserHook) -> TParserHook: ...



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause

import pathlib


class BuildBackendHookCaller:
    def __init__(self, source_dir: pathlib.Path, build_backend: str) -> None: ...
    def build_sdist(self, distdir: pathlib.Path) -> str: ...



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Parse a list of tags in the Tox configuration.

Inspired by https://github.com/tox-dev/tox-tags
"""

from __future__ import annotations

import importlib.metadata
import typing

import packaging.version as pver


if typing.TYPE_CHECKING:
    from typing import Final


def _get_tox_version() -> str | None:
    """Figure out which version of Tox is installed."""
    try:
        dist: Final = importlib.metadata.metadata("tox")
    except ModuleNotFoundError:
        return None

    return dist.get("Version")


_TOX_VERSION: Final = _get_tox_version()

if _TOX_VERSION is None:
    HAVE_MOD_TOX_3 = False
    HAVE_MOD_TOX_4 = False
else:
    HAVE_MOD_TOX_3 = pver.Version("3") <= pver.Version(_TOX_VERSION) < pver.Version("4")

    HAVE_MOD_TOX_4 = pver.Version("4") <= pver.Version(_TOX_VERSION) < pver.Version("5")


if HAVE_MOD_TOX_3:
    import tox
    import tox.config

    @tox.hookimpl
    def tox_addoption(parser: tox.config.Parser) -> None:
        """Parse a testenv's "tags" attribute as a list of lines."""
        parser.add_testenv_attribute(
            "tags",
            "line-list",
            "A list of tags describing this test environment",
            default=[],
        )


if HAVE_MOD_TOX_4:
    from typing import List

    import tox.plugin as t_plugin

    if typing.TYPE_CHECKING:
        import tox.config.sets as t_sets
        import tox.session.state as t_state

    @t_plugin.impl
    def tox_add_env_config(
        env_conf: t_sets.EnvConfigSet,
        state: t_state.State,  # noqa: ARG001
    ) -> None:
        """Parse a testenv's "tags" attribute as a list of lines."""
        env_conf.add_config(
            keys=["tags"],
            of_type=List[str],
            default=[],
            desc="A list of tags describing this test environment",
        )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Load the Tox configuration, look for our tags thing."""

from __future__ import annotations

import contextlib
import pathlib
import shutil
import subprocess
import sys
import tempfile
from typing import TYPE_CHECKING

import pytest
import utf8_locale

import tox_trivtags
import tox_trivtags.parse as ttt_parse


if sys.version_info >= (3, 11):
    import contextlib as contextlib_chdir
else:
    import contextlib_chdir

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from contextlib import AbstractContextManager
    from typing import Final


_EXPECTED: Final[dict[str, list[str]]] = {
    "format": ["check", "quick"],
    "reformat": ["format", "manual"],
    "unit-tests-no-tox": ["tests"],
    "unit-tests-tox-3": ["tests"],
    "unit-tests-tox-4": ["tests"],
    (".package" if tox_trivtags.HAVE_MOD_TOX_3 else ".pkg"): [],
    "t-single": ["something"],
    "t-several": ["all", "the", "things"],
    "t-special": ["So,", "how many", "$tags", 'is "too many",', "'eh\"?"],
}


def copy_and_adapt(srcdir: pathlib.Path, dstdir: pathlib.Path) -> None:
    """Copy some files over, adapt the tox.ini file."""
    src_tox = srcdir / "tox.ini"
    dst_tox = dstdir / "tox.ini"
    lines: Final = src_tox.read_text(encoding="UTF-8").splitlines()
    adapted: Final = [
        line.replace(" \\", "") for line in lines if not line.startswith("minversion")
    ]
    dst_tox.write_text("".join(line + "\n" for line in adapted), encoding="UTF-8")
    res: Final = subprocess.run(["diff", "-u", "--", src_tox, dst_tox], check=False)
    assert res.returncode != 0

    shutil.copytree(srcdir / "config", dstdir / "config")
    shutil.copytree(srcdir / "requirements", dstdir / "requirements")
    shutil.copy2(srcdir / "pyproject.toml", dstdir / "pyproject.toml")


@contextlib.contextmanager
def adapt_tox_cwd() -> Iterator[pathlib.Path]:
    """If using Tox 3.x, copy the files over and adapt them."""
    cwd: Final = pathlib.Path().absolute()
    if not tox_trivtags.HAVE_MOD_TOX_3:
        print("Not using Tox 3.x, no need to copy or adapt any files")
        yield cwd
        return

    with tempfile.TemporaryDirectory() as toxd_name:
        toxd = pathlib.Path(toxd_name).absolute()
        print(f"Adapting files for Tox 3.x from {cwd} to {toxd}")
        copy_and_adapt(cwd, toxd)
        with contextlib_chdir.chdir(toxd):
            yield toxd


@contextlib.contextmanager
def _cfg_filename_cwd() -> Iterator[pathlib.Path]:
    """No arguments, parse the tox.ini file in the current directory."""
    yield pathlib.Path("tox.ini")


@contextlib.contextmanager
def _cfg_filename_tempdir() -> Iterator[pathlib.Path]:
    """Create a temporary directory, enter it, pass `-c` with the original cwd."""
    cwd: Final = pathlib.Path().absolute()
    with tempfile.TemporaryDirectory() as tempd:
        print(f"Temporary directory: {tempd}; current directory: {cwd}")
        with contextlib_chdir.chdir(tempd):
            yield cwd / "tox.ini"


def _do_test_run_showconfig(filename: pathlib.Path) -> None:
    """Parse the `tox --showconfig` output."""
    u8env: Final = utf8_locale.UTF8Detect().detect().env
    print(f"Using {u8env['LC_ALL']} as a UTF-8-capable locale")

    envs: Final = ttt_parse.parse_showconfig(filename, env=u8env)
    print(f"Got some Tox config sections: {' '.join(sorted(envs))}")
    for envname, expected in _EXPECTED.items():
        print(f"- envname {envname!r} expected {expected!r}")
        assert envs[envname].tags == expected


@pytest.mark.parametrize("cfg_filename", [_cfg_filename_cwd, _cfg_filename_tempdir])
def test_run_showconfig(cfg_filename: Callable[[], AbstractContextManager[pathlib.Path]]) -> None:
    """Run `tox --showconfig` expecting tox.ini to be in the specified directory."""
    print()
    with adapt_tox_cwd(), cfg_filename() as filename:
        _do_test_run_showconfig(filename)
